/*
 * Copyright (C) by Klaas Freitag <freitag@owncloud.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 * for more details.
 */
#include "common/utility.h"
#include "account.h"
#include "simplesslerrorhandler.h"

namespace OCC {

bool SimpleSslErrorHandler::handleErrors(const QList<QSslError> &errors, const QSslConfiguration &, QList<QSslCertificate> *certs, OCC::AccountPtr)
{
    Q_ASSERT(certs);
    for (const auto &error : qAsConst(errors)) {
        certs->append(error.certificate());
    }
    return true;
}
}
